#!/usr/bin/env python
# coding: utf-8

# ### Modern Wage Dynamics Model
# version 1.0.0
#
# June 2022
# J M Applegate
#
# An implemetation of coupled production and allocation systems with efficiency wage theory of labour. 
# Multiple households provide labour to an aggregate firm to produce a single consumption good.
# The households choose how many hours of labour to provide and at what effort based on the offered wage
# and institutional considerations such as tax credits, unemployment benefits, and universal income.
#
# This version of the model is written in Python3 with a functional structure.
# Agent state variables are stored as an array per variable, where the array index indicates the relevant agent.
# Name experiments with date, parameters, scenario and run series.
# Run with a monthly time-scale.
# Includes benefits, minimum wage, EITCs and base subsistence needs.
#
# Version developed for workshopping at the 2022 Summer School titled `Mathematics of Complex Social Systems' at Zuse Institute, Berlin, July 9 - 17, 2022.
#
import numpy as np
import pandas as pd

from numpy.random import default_rng
from math import floor
from sklearn.preprocessing import MinMaxScaler
from sklearn.neural_network import MLPRegressor
from operator import itemgetter

import warnings
warnings.filterwarnings('ignore')

# series setup
from series_params import series_params
directory, series_name, verbose, seed, n_sets, sim_params_list, reps = series_params()
rng = default_rng(seed)

from initialisation_functions import *
from firm_functions import *
from household_functions import *
from market_functions import *

sim_labels = ['set', 'run', 'step', 'n', 'omega_max', 'H_max', 'A', 'gamma', 'mu', 'pct', 'e_min', 'EITC', 'upsilon', 'omega_min', 'sigma', 'S_N', 'd_max', 'omega_0', 'p_0', 'I', 'pi', 'total_pi', 'omega', 'p', 'e_hat', 'H_D', 'S_S', 'S_P', 'perceived_effort', 'household', 'd', 'omega_r', 'alpha', 'beta', 'e', 'H_N', 'H', 'H_W', 'm', 'S', 'S_C', 'income', 'tau', 'U', 'expenditure', 'N', 'H_S', 'H_M', 'S_D', 'S_M']
nn_labels= ['set', 'run', 'step', 'omega', 'effort']

sim_results = []
nn_results = []

print('This experiment consists of {} simulations.'.format(n_sets * reps))

#### main body of multi parameter set code
for s in range(n_sets):
    print('\nparameter set', s + 1)
    sim_params = sim_params_list[s]

    #unpack parameters
    t_max, omega_max, H_max, mu, pct, e_min, c1, c2, c3, c4, n, omega_0, A, gamma, EITC, upsilon, omega_min, sigma, S_N, p_0, d_max = itemgetter('t_max', 'omega_max', 'H_max', 'mu', 'pct', 'e_min', 'c1', 'c2', 'c3', 'c4', 'n', 'omega_0', 'A', 'gamma', 'EITC', 'upsilon', 'omega_min', 'sigma', 'S_N', 'p_0', 'd_max')(sim_params)

    for r in range(reps):
        if verbose: print('run: ', r + 1)
        elif r == 0: print('run: ', r + 1, end = ' ')
        else: print(r + 1,  end = ' ')

        t = 0
        # if verbose: print('iteration: ', t)
        # else: print('iteration: ', t, end = ' ')

        households = np.arange(0, n)

        # intitialise firm and household, initialise firm expectation models
        if verbose: print('Initial wage is', omega_0, ', initial price is', p_0, ', aggregate subsistance quantity is ', round(S_N * n, 3), 'and maximum labour is ', round(H_max * n, 3))
        I, pi, total_pi, omega, p, e_hat, H_D, S_P, S_S, perceived_effort = create_firm(omega_min, omega_0, p_0)
        d, alpha, beta, e, H_N, H, omega_r, H_W, m, S, S_C, income, tau, U, expenditure = create_households(n, H_max, d_max, upsilon, rng)
        N, H_S, H_M, S_M = 0, 0, 0, 0
        idx = rng.integers(0, n)
        omegas, omegas_scaled, omegas_train, efforts_train, omega_scaler, omega_memory, effort_memory, eta = initialise_eta(d[idx], omega_r[idx], omega_max, mu, rng)
        e_hat = min(eta.predict(omega_scaler.transform([[omega]]))[0], 1)
        if verbose: print('Initial e_hat is', round(e_hat, 3))
        #Find initial expected sugar demand
        H_N_init = tribute_hours(p, S_N, omega, m[idx], sigma, tau[idx], H_max)
        H_init = determine_hours(omega_r[idx], beta[idx], alpha[idx], H_max, H_N_init, omega, tau[idx], m[idx], sigma, p, S_N)
        S_D = n / p * (omega * H_init + m[idx] + tau[idx] + sigma)
        if verbose: print('Initial tribute hours are', H_N_init, ', initial hours supplied are', round(H_init, 3), 'and initial sugar demand is', round(S_D, 3))

        #save initial results
        step_results = [s, r, t,  n, omega_max, H_max, A, gamma, mu, pct, e_min, EITC, upsilon, omega_min, sigma, S_N, d_max, omega_0, p_0, I, pi, total_pi, omega, p, e_hat, H_D, S_S, S_P, e_hat, households, d, omega_r.copy(), alpha, beta, e.copy(), H_N.copy(), H.copy(), H_W, m.copy(), S.copy(), S_C, income.copy(), tau.copy(), U, expenditure.copy(), N, H_S, H_M, S_D, S_M]
        sim_results.append(step_results)

        # Save inital neural network
        nn_results.append([s, r, t, omegas, eta.predict(omegas_train)])
# 17 May this all checks out so far.

        # the simulation steps
        for t in range(1, t_max + 1):
            # if verbose: print('iteration: ', t)
            # else: print(t, end = ' ')

            #firm(s) determines hours
            if verbose: print('omega is ', round(omega, 3), ' and expected effort is ', round(e_hat, 3))
            N_opt, H_D, S_P = optimal_hours(S_D, A, gamma, e_hat, S_N, n)
            if verbose: print('optimal N is ', round(N_opt, 3), ' and hours demanded are ', round(H_D, 3))
            if verbose: print('planned sugar production is', round(S_P, 3))
            if verbose: print('price is ', round(p, 3))

            #households determine effort and hours supplied
            for i in households:
                H_N[i] = tribute_hours(p, S_N, omega, m[i], sigma, tau[i], H_max)
                omega_r[i] = upsilon / H[i] - tau[i] / H[i]
                e[i] = determine_effort(omega, d[i], omega_r[i], e_min)
                H[i] = determine_hours(omega_r[i], beta[i], alpha[i], H_max, H_N[i], omega, tau[i], m[i], sigma, p, S_N)
            if verbose: print('tribute hours are ', np.round(H_N, 3))
            if verbose: print('reservation wages are ', np.round(omega_r, 3))
            if verbose: print('hourly unemployment benefits are ', np.round(upsilon / H[i], 3))
            if verbose: print('EITC benefits are ', np.round(tau, 3))
            if verbose: print('wage sigma is', np.round(sigma / H[i], 3))
            if verbose: print('monetary holdings are', np.round(m, 3))
            if verbose: print('d values are ', np.round(d, 3))
            if verbose: print('household efforts at offered wage are ', np.round(e, 3))
            if verbose: print('mean effort is', round(np.mean(e), 3))
            if verbose: print('beta values are ', np.round(beta, 3))
            if verbose: print('household hours supplied are ', np.round(H, 3))

            #market aggregates hours supplied, determines market hours and effective effort
            H_S= np.sum(H)
            H_M = min(H_D, H_S)
            N, H_W = aggregate_effort(H_S, H_D, H, e)
            if verbose: print('hours supplied are ', round(H_S, 3), 'and hours purchased are', round(H_M, 3))
            if verbose: print('household hours worked are', np.round(H_W, 3))
            if verbose: print('household effective labour supplied is ', np.round(e * H_W, 3))
            if verbose: print('aggregate effort is ', np.round(N, 3))

            #firm produces sugar with effective labour
            S_S = produce_supply(A, N, gamma, I)
            if verbose: print('sugar supplied is ', round(S_S, 3), 'at a cost of', round(H_M * omega, 3))

            #households plan sugar consumption
            for i in households:
                S[i], income[i], tau[i] = plan_consumption(S_N, omega, upsilon, p, H_W[i], m[i], EITC, tau[i], sigma, c1, c2, c3, c4)
            if verbose: print('household income is', np.round(income, 3))
            if verbose: print('household demands are ', np.round(S, 3))

            #market aggregates sugar demand and determines sugar sold
            S_D = np.sum(S)
            S_M = min(S_S, S_D)
            S_C = sell_sugar(S_S, S_D, S)
            if verbose: print('aggregate sugar demand is ', round(S_D, 3), ' sugar sold is ', round(S_M, 3))
            if verbose: print('household sugar comsumption quantities are ', np.round(S_C, 3))

            #calculate firm profit
            pi = profit(omega, p, S_M, H_M)
            total_pi += pi
            if verbose: print('firm profit is ', np.round(pi, 3), 'and total profit is', round(total_pi,3))

            #calculate household utility
            U = calculate_utility(H_max, H_W, S_C, S_N, alpha, beta)
            if verbose: print('household utilites are ', np.round(U, 3))

            #households adjust ledgers
            for i in households:
                m[i], expenditure[i] = update_ledger(income[i], tau[i], sigma, upsilon, p, S_C[i], m[i])
            if verbose: print('household spent ', np.round(expenditure, 3))
            if verbose: print('households monetary holdings are ', np.round(m, 3))
            if verbose: print('total profit is', round(total_pi, 3), 'and total household holdings are', np.round(sum(m), 3))

            #firm adjusts inventory
            I = adjust_inventory(S_S, S_M, I)
            if verbose: print('firm inventory is ', round(I, 3))

            #firm updates expectations
            perceived_effort, omega_memory, effort_memory, eta = update_eta(eta, omega, S_S, H_M, A, gamma, omega_memory, effort_memory, omega_scaler)
            if verbose: print('perceived effort is', round(perceived_effort, 3))

            #Save eta state to frame
            nn_results.append([s, r, t, omegas, eta.predict(omegas_train)])

            # firm raises or lowers wage
            omega, p, e_hat = update_wage_price(omega, pct, omega_min, p, pi, H_S, H_D, S_S, S_D, omega_scaler, eta)

            #write step state to frame
            step_results = [s, r, t,  n, omega_max, H_max, A, gamma, mu, pct, e_min, EITC, upsilon, omega_min, sigma, S_N, d_max, omega_0, p_0, I, pi, total_pi, omega, p, e_hat, H_D, S_S, S_P, perceived_effort, households, d, omega_r.copy(), alpha, beta, e.copy(), H_N.copy(), H.copy(), H_W, m.copy(), S.copy(), S_C, income.copy(), tau.copy(), U, expenditure.copy(), N, H_S, H_M, S_D, S_M]
            sim_results.append(step_results)

print('\nFinished series.')

#transform list of lists into dataframes and expand subarrays
sim_results_frame = pd.DataFrame(sim_results, columns = sim_labels)
sim_results_long = sim_results_frame.explode(['household', 'd', 'omega_r', 'alpha', 'beta', 'e', 'H_N', 'H', 'H_W', 'm', 'S', 'S_C', 'income', 'tau', 'U', 'expenditure'])
nn_results_frame = pd.DataFrame(nn_results, columns = nn_labels)
nn_results_long = nn_results_frame.explode(['omega', 'effort'])

#### Save expanded dataframes to directory as csv files.
print('Writing to files.')
prefix = directory + 'series_' + series_name
#nn_results_long.to_csv(prefix + 'NN.csv', index = False)
sim_results_long.to_csv(prefix + '.csv', index = False)
print('Series results written to', prefix + '.csv')
